@@ -34,7 +34,7 @@
 #include "nm-core-internal.h"
 #include "nm-simple-connection.h"
 
-#include "nmdbus-vpn-plugin.h"
+#include "introspection/org.freedesktop.NetworkManager.VPN.Plugin.h"
 
 #define NM_VPN_SERVICE_PLUGIN_QUIT_TIMER    180
 
@@ -225,6 +225,7 @@ nm_vpn_service_plugin_disconnect (NMVpnServicePlugin *plugin, GError **err)
 		break;
 	case NM_VPN_SERVICE_STATE_INIT:
 		ret = TRUE;
+		nm_vpn_service_plugin_set_state (plugin, NM_VPN_SERVICE_STATE_STOPPED);
 		break;
 
 	default:
@@ -554,7 +555,7 @@ impl_vpn_service_plugin_connect_interactive (NMVpnServicePlugin *plugin,
 	_connect_generic (plugin, context, connection, details);
 }
 
-/***************************************************************/
+/*****************************************************************************/
 
 static void
 impl_vpn_service_plugin_need_secrets (NMVpnServicePlugin *plugin,
@@ -702,7 +703,7 @@ nm_vpn_service_plugin_secrets_required (NMVpnServicePlugin *plugin,
 	nmdbus_vpn_plugin_emit_secrets_required (priv->dbus_vpn_service_plugin, message, hints);
 }
 
-/***************************************************************/
+/*****************************************************************************/
 
 #define DATA_KEY_TAG "DATA_KEY="
 #define DATA_VAL_TAG "DATA_VAL="
@@ -868,7 +869,7 @@ nm_vpn_service_plugin_get_secret_flags (GHashTable *data,
 	return success;
 }
 
-/***************************************************************/
+/*****************************************************************************/
 
 static void
 impl_vpn_service_plugin_disconnect (NMVpnServicePlugin *plugin,
@@ -923,7 +924,7 @@ impl_vpn_service_plugin_set_failure (NMVpnServicePlugin *plugin,
 	g_dbus_method_invocation_return_value (context, NULL);
 }
 
-/*********************************************************************/
+/*****************************************************************************/
 
 static void
 sigterm_handler (int signum)
@@ -945,7 +946,7 @@ setup_unix_signal_handler (void)
 	sigaction (SIGTERM, &action, NULL);
 }
 
-/*********************************************************************/
+/*****************************************************************************/
 
 static void
 one_plugin_destroyed (gpointer data,
@@ -994,19 +995,6 @@ init_sync (GInitable *initable, GCancellable *cancellable, GError **error)
 	if (!proxy)
 		goto out;
 
-	ret = g_dbus_proxy_call_sync (proxy,
-	                              "RequestName",
-	                              g_variant_new ("(su)", priv->dbus_service_name, 0),
-	                              G_DBUS_CALL_FLAGS_NONE, -1,
-	                              cancellable, error);
-	g_object_unref (proxy);
-	if (!ret) {
-		if (error && *error)
-			g_dbus_error_strip_remote_error (*error);
-		goto out;
-	}
-	g_variant_unref (ret);
-
 	priv->dbus_vpn_service_plugin = nmdbus_vpn_plugin_skeleton_new ();
 	if (!g_dbus_interface_skeleton_export (G_DBUS_INTERFACE_SKELETON (priv->dbus_vpn_service_plugin),
 	                                       connection,
@@ -1030,6 +1018,19 @@ init_sync (GInitable *initable, GCancellable *cancellable, GError **error)
 	nm_vpn_service_plugin_set_connection (plugin, connection);
 	nm_vpn_service_plugin_set_state (plugin, NM_VPN_SERVICE_STATE_INIT);
 
+	ret = g_dbus_proxy_call_sync (proxy,
+	                              "RequestName",
+	                              g_variant_new ("(su)", priv->dbus_service_name, 0),
+	                              G_DBUS_CALL_FLAGS_NONE, -1,
+	                              cancellable, error);
+	g_object_unref (proxy);
+	if (!ret) {
+		if (error && *error)
+			g_dbus_error_strip_remote_error (*error);
+		goto out;
+	}
+	g_variant_unref (ret);
+
 	success = TRUE;
 
  out:
@@ -1306,3 +1307,11 @@ nm_vpn_service_plugin_initable_iface_init (GInitableIface *iface)
 {
 	iface->init = init_sync;
 }
+
+/*****************************************************************************/
+
+/* this header is intended to be copied to users of nm_vpn_editor_plugin_call(),
+ * to simplify invocation of generic functions. Include it here, to complile
+ * the code. */
+#include "nm-utils/nm-vpn-editor-plugin-call.h"
+
